\AMaze3D.xpl	18-Aug-2009	Loren Blaney	loren.blaney@idcomm.com
\Displays randomly generated mazes that can be manipulated in 3D
\Compile with 32-bit XPL0 (XPLPX, xpx.bat)
\
\ A right-hand coordinate system is used:
\
\				|   Z  (into screen)
\				|  /
\				| /
\				|/
\			--------+-------- X
\			       /|
\			      / |
\			     /	|
\				|
\				Y

include	c:\cxpl\codes;		\intrinsic code definitions (including Paint)

def	VidMode=$105,		\video mode: 256-color graphics
	ScrW=1024, ScrH=768,	\screen width and height, in pixels
	ImW=ScrH, ImH=ScrH;	\image width and height in pixels (a square)

int	CpuReg;			\address of CPU register array (from GetReg)
char	Image(ImW*ImH);		\image buffer, gets copied to the graphic screen

def	Bel=$07, CR=$0D, Esc=$1B;				\control chars
def	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;	\scan codes
def	Red=$28, Green=$78, Blue=$38, Magenta=$58;



func	WExtend(N);		\Sign-extend 16-bit word to 32-bit double word
int	N;
return if N & $8000 then N ! $FFFF0000 else N & $0000FFFF;



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
[CpuReg(0):= AX;
 CpuReg(1):= BX;
 CpuReg(2):= CX;
 CpuReg(3):= DX;
 CpuReg(6):= BP;
 CpuReg(9):= DS;
 CpuReg(11):= ES;
 SoftInt(Int);
 return WExtend(CpuReg(0));	\return value in AX register (as 32 bits)
];	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\This is a low-level routine with no echo,
[				\ no Ctrl+C, and no cursor
SC:= CallInt($16, $0000);	\BIOS function $00
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
];	\GetKey



func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
[				\Pointer is set to center of screen but hidden
CallInt($21, $3533);		\Make sure mouse vector $33 points to something
if WExtend(CpuReg(1))=0 & WExtend(CpuReg(11))=0 then return false;
CallInt($33, $21);		\reset mouse for Windows
CallInt($33, $0B);		\read motion counters to reset them for Windows
return CallInt($33, $0000);	\h/w reset mouse and return status
];	\OpenMouse

\-------------------------------------------------------------------------------

proc	CrossProd(V1, V2, V3);	\Calculate the cross product of two 3D vectors
real	V1, V2, V3;		\V3:= V1 x V2
[V3(0):= V1(1)*V2(2) - V1(2)*V2(1);
 V3(1):= V1(2)*V2(0) - V1(0)*V2(2);
 V3(2):= V1(0)*V2(1) - V1(1)*V2(0);
];	\CrossProd



proc	Rotate(V, R, P, W); \Rotate vector in 3 dimensions
real	V,		\3D vector, a 3-element array with X, Y and Z components
	R, P, W;	\Roll, Pitch and yaW (radians)
real	SW, SP, SR, CW, CP, CR, T;
[SR:= Sin(R);   SP:= Sin(P);   SW:= Sin(W);
 CR:= Cos(R);   CP:= Cos(P);   CW:= Cos(W);

\Rotate about X axis (Roll):
T:= V(1)*CR - V(2)*SR;		\		       	 
V(2):= V(2)*CR + V(1)*SR;	\		      1   0   0 
V(1):= T;			\ [X Y Z]:= [X Y Z] * 0   CR  SR
				\		      0  -SR  CR
				\		       	 
\Rotate about Y axis (Pitch):
T:= V(0)*CP + V(2)*SP;		\		       	 
V(2):= V(2)*CP - V(0)*SP;	\		      CP  0  -SP
V(0):= T;			\ [X Y Z]:= [X Y Z] * 0   1   0 
				\		      SP  0   CP
				\		       	 
\Rotate about Z axis (yaW):
T:= V(0)*CW - V(1)*SW;		\		       	  
V(1):= V(1)*CW + V(0)*SW;	\		       CW  SW	0 
V(0):= T;			\ [X Y Z]:= [X Y Z] * -SW  CW	0 
				\		       0   0	1 
];	\Rotate					       	  



proc	EraseImage;		\Erase the Image array
int	Im, I;
[Im:= Image;			\erase 4 bytes at a time, for speed
for I:= 0, (ImW*ImH)>>2-1 do Im(I):= 0;
];

\-------------------------------------------------------------------------------

int	PenX, PenY;		\pen coordinates (pixels) for drawing onto Image



proc	MovePen(X, Y);		\Position drawing pen
int	X, Y;
[PenX:= X;  PenY:= Y];



proc	DrawLine(X, Y, C); \Draw line on Image from PenX,PenY to X,Y in color C
int	X, Y, C;	\coordinates of end-of-line, color
int	DX, DY,		\delta X, delta Y
	IX, IY,		\plotting increments along X and Y axes
	SY,		\flag: step along Y direction
	D,		\decision variable: move straight or diagonally
	E, NE,		\East, Northeast (straight or diagonal)
	T, I;		\scratch
[DX:= X - PenX;				\delta X
if DX >= 0 then				\if positive then increment
	IX:= 1
else	[DX:= -DX;   IX:= -1];		\else absolute value and decrement

DY:= Y - PenY;				\same for Y direction
if DY >= 0 then
	IY:= 1
else	[DY:= -DY;   IY:= -1];

if DX >= DY then			\step along X direction
	SY:= false
else	[SY:= true;			\step along Y direction
	T:= DX;   DX:= DY;   DY:= T;	\swap DX and DY
	];

D:= 2*DY - DX;				\set up decision variable stuff
E:= 2*DY;
NE:= 2*(DY-DX);

if PenX>=0 & PenX<ImW & PenY>=0 & PenY<ImH then	\clip
	Image(PenX + PenY*ImW):= C;	\plot initial point
for I:= 1, DX do
	[if D < 0 then
		[D:= D + E;
		if SY then PenY:= PenY + IY
		      else PenX:= PenX + IX;
		]
	else	[D:= D + NE;		\move diagonally
		PenX:= PenX + IX;
		PenY:= PenY + IY;
		];
	if PenX>=0 & PenX<ImW & PenY>=0 & PenY<ImH then	\clip
		Image(PenX + PenY*ImW):= C;
	];
];	\DrawLine

\-------------------------------------------------------------------------------

proc	DrawTri(Tri);		\Draw a filled triangle onto Image array
int	Tri;			\array of 3 points(x,y) and one 8-bit color
int	X0,Y0, X1,Y1, X2,Y2, Color;
int	X, Y, T;
int	Top, Bot,		\top and bottom Y coordinates
	YImW;			\Y coordinate times Image width
int	Edge1, Edge2;		\X coordinates along edges of triangles
real	REdge1, REdge2, Slope1, Slope2;
char	Img;

	proc	DrawHalf;	\(optimized procedure with no local variables)
	[if Bot > ImH then Bot:= ImH-1;			\clip to bottom of Image
	YImW:= Top*ImW;
	for Y:= Top, Bot-1 do				\from top to bottom...
		[if Y >= 0 then				\clip to top of Image
			[Edge1:= fix(REdge1);  Edge2:= fix(REdge2);
			if Edge1 > Edge2 then		\order left to right
				[T:= Edge1;  Edge1:= Edge2;  Edge2:= T];
			if Edge1 < 0 then Edge1:= 0;	\clip to left side
			if Edge2 > ImW then Edge2:= ImW;\clip to right side
			for Img:= Image+Edge1+YImW, Image+Edge2-1+YImW do
				Img(0):= Color;		\fill scanline
			];
		REdge1:= REdge1 + Slope1;		\adj X coords according
		REdge2:= REdge2 + Slope2;		\ to slopes
		YImW:= YImW + ImW;
		];
	];	\DrawHalf

[	\DrawTri
X0:= Tri(0,0);			\get vertices
X1:= Tri(1,0);
X2:= Tri(2,0);

if X0>=ImW & X1>=ImW & X2>=ImW then return;	\clip entire triangle
if X0<0 & X1<0 & X2<0 then return;

Y0:= Tri(0,1);
Y1:= Tri(1,1);
Y2:= Tri(2,1);

if Y0 > Y1 then			\sort vertices from top to bottom (Y0<=Y1<=Y2)
	[T:= Y1;  Y1:= Y0;  Y0:= T;
	 T:= X1;  X1:= X0;  X0:= T;
	];
if Y0 > Y2 then
	[T:= Y2;  Y2:= Y0;  Y0:= T;
	 T:= X2;  X2:= X0;  X0:= T;
	];
if Y0 >= ImH then return;	\clip
if Y1 > Y2 then
	[T:= Y2;  Y2:= Y1;  Y1:= T;
	 T:= X2;  X2:= X1;  X1:= T;
	];
if Y2 < 0 then return;		\clip

if Y1 # Y0 then Slope1:= float(X1-X0) / float(Y1-Y0);
if Y2 # Y0 then Slope2:= float(X2-X0) / float(Y2-Y0);
REdge1:= float(X0);  REdge2:= REdge1;
Top:= Y0;  Bot:= Y1;
Color:= Tri(3);
DrawHalf;

if Y2 # Y1 then Slope1:= float(X2-X1) / float(Y2-Y1);
REdge1:= float(X1);
Top:= Y1;  Bot:= Y2;
DrawHalf;
];	\DrawTri



proc	DrawQuad(Quad);		\Draw a filled quadrilateral onto Image array
int	Quad;			\array of 4 points(x,y) and one 8-bit color
int	Tri(4,2);
[Tri(3):= Quad(4);		\color

Tri(0,0):= Quad(0,0);
Tri(0,1):= Quad(0,1);
Tri(1,0):= Quad(1,0);
Tri(1,1):= Quad(1,1);
Tri(2,0):= Quad(2,0);
Tri(2,1):= Quad(2,1);
DrawTri(Tri);

Tri(0,0):= Quad(0,0);
Tri(0,1):= Quad(0,1);
Tri(1,0):= Quad(2,0);
Tri(1,1):= Quad(2,1);
Tri(2,0):= Quad(3,0);
Tri(2,1):= Quad(3,1);
DrawTri(Tri);
];	\DrawQuad

\###############################################################################

proc	MoveMaze;	\Move and draw the maze
def	Max = 33;	\maximum limits +1 for Cols and Rows
int	Cols, Rows,	\number of columns and rows in the maze
	Facets,		\number of facets (walls and ceilings)
	FPS,		\frames per second, display update rate
	Sec,		\current second of time (from real-time clock)
	SecOld,		\previous second of time
	Solve;		\flag: solve the maze
int	Cell(Max+1, Max+1, 3);	\cells (plus right and bottom borders)
def \Cell\ LeftWall, Ceiling, Connected;	\attributes of each cell
real	Zoom,		\magnification factor
	Vertex(Max,Max,2,3), \coordinates of (wall and ceiling) vertices
	RVertex(Max,Max,2,3), \coordinates of rotated vertices
	Ref1(3), Ref2(3), Ref3(3); \reference frame (unit vectors)
int	Offset(3),	 \offset position to add to vertices (X,Y,Z)
	RCell(Max,Max,2,3), \rotated coordinates of vertices
	RCellZZ(Max*Max*2), \decreasing Z coordinates of rotated maze
	RCellXY(Max*Max*2); \corresponding X and Y indexes for sorted Z coords



proc	Sort(AZ, AI, N);	\Sort arrays by their decreasing Z coordinates
int	AZ,	\address of array of Z coordinates
	AC,	\address of array of X and Y coordinates
	N;	\number of elements in array (size)
int	I, J, Gap, JG, T;
[Gap:= N>>1;			\shell sort
while Gap > 0 do
	[for I:= Gap, N-1 do
		[J:= I - Gap;
		loop	[JG:= J + Gap;
			if AZ(J) >= AZ(JG) then quit;
			T:= AZ(J);  AZ(J):= AZ(JG);  AZ(JG):= T;	\swap
			T:= AC(J);  AC(J):= AC(JG);  AC(JG):= T;
			J:= J - Gap;
			if J < 0 then quit;
			];
		];
	Gap:= Gap>>1;
	];
];	\Sort



proc	DepthSort;		\Sort facets by decreasing depth
\Find the Z coordinate (*2) of the approximate center of each facet. Save the
\ result in array RCellZZ. Array RCellXY saves X and Y coordinates for the
\ corresponding facet. (Z increases going into the screen.)
\Inputs: Cols, Rows, RCell.
\Outputs: RCellZZ, RCellXY, Facets.
int	X, Y;
[Facets:= 0;
for Y:= 0, Rows do
    for X:= 0, Cols do
	[if Y # Rows then
		[RCellZZ(Facets):= RCell(X,Y,0,2) + RCell(X,Y+1,1,2);	\wall
		 RCellXY(Facets):= X + Y<<8 + 0<<16;
		 Facets:= Facets+1;
		];
	if X # Cols then
		[RCellZZ(Facets):= RCell(X,Y,0,2) + RCell(X+1,Y,1,2);	\ceiling
		 RCellXY(Facets):= X + Y<<8 + 1<<16;
		 Facets:= Facets+1;
		];
	];
Sort(RCellZZ, RCellXY, Facets);
];	\DepthSort

\-------------------------------------------------------------------------------

proc	DrawWall(Wall);		\Draw a maze wall (or ceiling)
int	Wall;			\array of 4 points(x,y) and one 8-bit color
int	I, Quad(5,2);		\ditto
def	CX=ImW/2, CY=ImH/2;	\center coordinate
[for I:= 0, 3 do
	[Quad(I,0):= Wall(I,0) + CX;
	 Quad(I,1):= Wall(I,1) + CY;
	];
Quad(4):= Wall(4);		\color
DrawQuad(Quad);

MovePen(Quad(0,0), Quad(0,1));	\outline wall, so it stands out from other walls
for I:= 1, 4 do
	DrawLine(Quad(I&3,0), Quad(I&3,1), Magenta);
];	\DrawWall



proc	DrawFloor;		\Draw the floor for the entire maze
int	X, Y;
int	Floor(5,2);
def	CX=ImW/2, CY=ImH/2;	\center coordinate
[for Y:= 0, Rows-1 do
    for X:= 0, Cols-1 do
	[Floor(0,0):= RCell(X,   Y,   1, 0) + CX;
	 Floor(0,1):= RCell(X,   Y,   1, 1) + CY;
	 Floor(1,0):= RCell(X+1, Y,   1, 0) + CX;
	 Floor(1,1):= RCell(X+1, Y,   1, 1) + CY;
	 Floor(2,0):= RCell(X+1, Y+1, 1, 0) + CX;
	 Floor(2,1):= RCell(X+1, Y+1, 1, 1) + CY;
	 Floor(3,0):= RCell(X,   Y+1, 1, 0) + CX;
	 Floor(3,1):= RCell(X,   Y+1, 1, 1) + CY;
	 Floor(4):= Green;
	 DrawQuad(Floor);
	];
];	\DrawFloor

\-------------------------------------------------------------------------------

proc	DrawMaze;		\Draw the maze (onto Image)
\Input: Cols, Rows, RCell, Facets, RCellXY
int	X, Y, C, F;		\indexes
int	Wall(5,2);
[DrawFloor;			\(it's always on the bottom, farthest way)

\Draw walls starting with the farthest one (the one with the greatest Z)
DepthSort;
for F:= 0, Facets-1 do
	[X:= RCellXY(F) & $00FF;				\unpack indexes
	Y:= RCellXY(F) >> 8 & $00FF;
	C:= RCellXY(F) >> 16;

	if C=0 & Cell(X,Y,LeftWall) then
		[Wall(0,0):= RCell(X,   Y, 0, 0);		\wall
		 Wall(0,1):= RCell(X,   Y, 0, 1);
		 Wall(1,0):= RCell(X, Y+1, 0, 0);
		 Wall(1,1):= RCell(X, Y+1, 0, 1);
		 Wall(2,0):= RCell(X, Y+1, 1, 0);
		 Wall(2,1):= RCell(X, Y+1, 1, 1);
		 Wall(3,0):= RCell(X,   Y, 1, 0);
		 Wall(3,1):= RCell(X,   Y, 1, 1);
		 Wall(4):= Red;
		 DrawWall(Wall);
		]
	else if C=1 & Cell(X,Y,Ceiling) then
		[Wall(0,0):= RCell(X,   Y, 0, 0);		\ceiling
		 Wall(0,1):= RCell(X,   Y, 0, 1);
		 Wall(1,0):= RCell(X+1, Y, 0, 0);
		 Wall(1,1):= RCell(X+1, Y, 0, 1);
		 Wall(2,0):= RCell(X+1, Y, 1, 0);
		 Wall(2,1):= RCell(X+1, Y, 1, 1);
		 Wall(3,0):= RCell(X,   Y, 1, 0);
		 Wall(3,1):= RCell(X,   Y, 1, 1);
		 Wall(4):= Blue;
		 DrawWall(Wall);
		];
	];
];	\DrawMaze

\-------------------------------------------------------------------------------

proc	SolveMaze;		\Use dead-end-filling to show path thru maze
int	X, Y, Ctr;
[Solve:= false;			\assume it's done solving
for Y:= 0, Rows-1 do		\scan through entire maze
    for X:= 0, Cols-1 do
	[Ctr:= 0;		\count number of walls surounding a cell
	if Cell(X, Y, LeftWall) then Ctr:= Ctr+1;
	if Cell(X, Y, Ceiling) then Ctr:= Ctr+1;
	if Cell(X+1, Y, LeftWall) then Ctr:= Ctr+1;
	if Cell(X, Y+1, Ceiling) then Ctr:= Ctr+1;

	if Ctr = 3 then		\it's a dead end; wall it off
		[Cell(X, Y, LeftWall):= true;
		 Cell(X, Y, Ceiling):= true;
		 Cell(X+1, Y, LeftWall):= true;
		 Cell(X, Y+1, Ceiling):= true;
		 Solve:= true;	\nope, still working on solution
		];
	];
];	\SolveMaze

\-------------------------------------------------------------------------------

proc	RotateMaze;		\Rotate, pan, and zoom maze with the mouse
\Inputs: Vertex, Cols, Rows, Zoom, Ref1, Ref2.
\Outputs: RCell, Ref1, Ref2.
int	X, Y, Z, D;		\indexes
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
real	DX, DY, DZ, Dist;
[AngY:= 0.0;
 AngX:= 0.0;
 AngZ:= 0.0;

CallInt($33, $0003);
case CpuReg(1) & $03 of
  $02:	[	\right button
	CallInt($33, $0B);		\read mouse motion counters
	Offset(0):= Offset(0) + WExtend(CpuReg(2));	\pan motion
	Offset(1):= Offset(1) + WExtend(CpuReg(3));
	];
  $03:	[	\both buttons
	CallInt($33, $0B);		\read mouse motion counters
	Zoom:= Zoom + Float(-WExtend(CpuReg(3)))/1000.0;
	]
other	[CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Y axis
	AngY:= Float(-WExtend(CpuReg(2))) / 500.0;
	\Mouse Y motion rotates object about X axis
	AngX:= Float( WExtend(CpuReg(3))) / 500.0;
	];
\Rotate reference frame according to changes in mouse position.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2

\Rotate each vertex in the maze so that it aligns with the reference frame.
\Multiply each point at its initial position (Po) by the rotation matrix to get
\ its rotated position:
\					     				 
\		     	 		     Ref1(0)  Ref1(1)  Ref1(2) 
\	 Px  Py  Pz  =  Pxo	Pyo  Pzo  *  Ref2(0)  Ref2(1)  Ref2(2) 
\		     	 		     Ref3(0)  Ref3(1)  Ref3(2) 
\					     				 
\
for D:= 0, 3-1 do			\for all dimensions (X,Y,Z)...
  for Z:= 0, 1 do			\for near and far (Z axis)...
    for Y:= 0, Rows do			\for all of the rows+1...
      for X:= 0, Cols do		\for all of the columns+1...
	RVertex(X,Y,Z,D):= Vertex(X,Y,Z,0)*Ref1(D) + Vertex(X,Y,Z,1)*Ref2(D) +
			   Vertex(X,Y,Z,2)*Ref3(D);

\Add perspective, to make closer objects appear bigger; also add pan and zoom
for Z:= 0, 1 do
  for Y:= 0, Rows do
    for X:= 0, Cols do
       [DX:= RVertex(X,Y,Z,0);
	DY:= RVertex(X,Y,Z,1);
	DZ:= RVertex(X,Y,Z,2) + 1500.0;		\add distance from viewer
	Dist:= sqrt(DX*DX + DY*DY + DZ*DZ) / 1500.0 / Zoom;
	RCell(X,Y,Z,0):= fix(RVertex(X,Y,Z,0)/Dist) + Offset(0);
	RCell(X,Y,Z,1):= fix(RVertex(X,Y,Z,1)/Dist) + Offset(1);
	RCell(X,Y,Z,2):= fix(RVertex(X,Y,Z,2));
       ];
];	\RotateMaze

\-------------------------------------------------------------------------------

proc	MakeMaze;	\Randomly generate maze and set up array of coordinates
\A 2D maze is defined in the X,Y plane and then extruded into the Z direction
\Inputs: Cols, Rows
\Outputs: Cell, Vertex



proc	ConnectFrom(X, Y);	\Connect cells starting from cell X,Y
int	X, Y;
int	Dir, Dir0;
[Cell(X, Y, Connected):= true;	\mark current cell as connected
Dir:= Ran(4);			\randomly choose a direction
Dir0:= Dir;			\save this initial direction
repeat	case Dir of		\try to connect to cell at Dir
	  0: if X+1<Cols & not Cell(X+1, Y, Connected) then	\go right
		[Cell(X+1, Y, LeftWall):= false; ConnectFrom(X+1, Y)];
	  1: if X-1>=0 & not Cell(X-1, Y, Connected) then	\go left
		[Cell(X, Y, LeftWall):= false;   ConnectFrom(X-1, Y)];
	  2: if Y+1<Rows & not Cell(X, Y+1, Connected) then	\go down
		[Cell(X, Y+1, Ceiling):= false;  ConnectFrom(X, Y+1)];
	  3: if Y-1>=0 & not Cell(X, Y-1, Connected) then	\go up
		[Cell(X, Y, Ceiling):= false;    ConnectFrom(X, Y-1)]
	other	[];		\(never occurs)
	Dir:= Dir+1 & $03;	\next direction
until	Dir = Dir0;
]; \ConnectFrom



proc	GenMaze;		\Generate random maze
int X, Y;
[for Y:= 0, Rows do
    for X:= 0, Cols do
	[Cell(X, Y, LeftWall):= true;		\start with all walls and
	 Cell(X, Y, Ceiling):= true;		\ ceilings in place
	 Cell(X, Y, Connected):= false;		\ and all cells disconnected
	];
Cell(0, 0, LeftWall):= false;			\make left and right doorways
Cell(Cols, Rows-1, LeftWall):= false;

ConnectFrom(Ran(Cols), Ran(Rows));		\randomly pick a starting cell
];	\GenMaze



def	CeilingLength = 30.0,			\wall dimensions
	WallLength = CeilingLength,
	WallHeight = 30.0;
int     X, Y, Z;	\indexes
real	CX, CY;		\offset from upper-left corner to center of maze
[	\MakeMaze
GenMaze;

\Create an array of 3D coordinates for the maze centered at 0,0,0
CX:= float(Cols) * CeilingLength / 2.0;
CY:= float(Rows) * WallLength / 2.0;
for Z:= 0, 1 do
  for Y:= 0, Rows do
    for X:= 0, Cols do
	[Vertex(X,Y,Z,0):= CeilingLength * float(X) - CX;
	 Vertex(X,Y,Z,1):= WallLength * float(Y) - CY;
	 Vertex(X,Y,Z,2):= WallHeight * float(Z);
	];
];	\MakeMaze

\-------------------------------------------------------------------------------

proc	Reset;			\Initialize variables to starting configuration
[Zoom:= 1.0;			\magnification factor
Offset(0):= 0;   Offset(1):= 0;   Offset(2):= 0;

Ref1(0):= 1.0;   Ref1(1):= 0.0;   Ref1(2):= 0.0;	\tilt reference frame 30
Ref2(0):= 0.0;   Ref2(1):= sqrt(3.0)/2.0;   Ref2(2):= -0.5;

MakeMaze;
Solve:= false;
];	\Reset



proc	NumOut(Line, Num);	\Display 3-digit right-justified positive number
int	Line, Num;		\ at the beginning of the specified Line
[Cursor(0, Line);
if Num < 100 then ChOut(6, ^ );
if Num < 10  then ChOut(6, ^ );
IntOut(6, Num);
];	\NumOut

\-------------------------------------------------------------------------------

[	\MoveMaze
Cols:= 20;  Rows:= 12;
FPS:= 0;
Reset;

Cursor(0, 0);			\(device 6 always works for VESA graphic modes)
Text(6, "Mouse rotates maze
Right button pans
Both buttons zoom


    Columns  [^[] [^Z]

    Rows     [^Y] [^X]

    Frames/second

    Solve    [S]

    Reset    [Enter]

    Exit     [Esc]");

loop	[NumOut(5, Cols);
	 NumOut(7, Rows);

	CallInt($1A, $0200);	\get current time in seconds
	Sec:= CpuReg(3)>>8 & $FF;
	if Sec = SecOld then	\if same as before then
		FPS:= FPS+1	\ bump FPS counter
	else	[SecOld:= Sec;	\save old seconds
		NumOut(9, FPS);\display FPS counter (once per second)
		FPS:= 0;	\reset FPS counter
		];

	if Solve then SolveMaze;

	RotateMaze;
	EraseImage;
	DrawMaze;
	Paint(ScrW-ImW, 0, ImW, ImH, Image, ImW);     \copy Image onto screen

	if KeyHit then
		case GetKey of
		  -LtArrow:
			if Cols > 2 then
				[Cols:= Cols-1;  MakeMaze];
		  -RtArrow:
			if Cols < Max-1 then
				[Cols:= Cols+1;  MakeMaze];
		  -UpArrow:
			if Rows < Max-1 then
				[Rows:= Rows+1;  MakeMaze];
		  -DnArrow:
			if Rows > 2 then
				[Rows:= Rows-1;  MakeMaze];
		  ^S, ^s:	Solve:= true;
		  CR:	Reset;
		  Esc:	quit
		other	ChOut(0, Bel);
	];
];	\MoveMaze

\###############################################################################

[	\Main
CpuReg:= GetReg;		\get address of array to access CPU registers
SetVid(VidMode);		\(in WinXP this affects mouse motion counters)
if not OpenMouse then [SetVid(3);  Text(0, "A mouse is required");  exit];
MoveMaze;
SetVid(3);			\restore normal text mode (for DOS)
];	\Main
