\AMaze3D2.xpl	18-Aug-2009	Loren Blaney	loren.blaney@idcomm.com
\Displays randomly generated spherical mazes that can be manipulated in 3D
\Compile with 32-bit XPL0 (XPLPX, xpx.bat)

include AMaze3D2.inc;		\fundamental support routines

\########################## Display and Move the Maze ##########################

proc	MoveMaze;		\Move and draw the maze

def	LevelsMax = 4,		\0	1	2	3	4    F+V = E+2
	FloorsMax = 5120,	\20	80	320	1280	5120 F
	NodesMax = 2562,	\12	42	162	642	2562 V
	WallsMax = 7680;	\30	120	480	1920	7680 E

int	Levels,			\current number of levels of breakdown
	Floors,			\corresponding number of floors (facets)
	Nodes,			\corresponding number of nodes (vertices)
	Walls;			\number of walls to draw (edges)
real	GeoVert(NodesMax,3);	\3D coords of points (vertexes) for each node
int	GeoNode(NodesMax,6),	\array of lists giving node inner connections
	GeoNodeInx(NodesMax),	\indexes into each list
	GeoPath(NodesMax,6);	\boolean: pathways leading out of each node
def	Phi = (1.0+sqrt(5.0))/2.0; \golden ratio

int	FPS,			\frames per second, display update rate
	Sec,			\current second of time (from real-time clock)
	SecOld,			\previous second of time
	Solve;			\flag: solve the maze
real	Zoom,			\magnification factor
	Ambient,		\intensity of ambient light
	Light(3),		\directional light (unit vector)
	RotVert(NodesMax,3),	\coordinates of rotated vertices
	WallVert(WallsMax,4,3),	\maximum possible walls, 4 corners, 3 dimensions
	WallNorm(WallsMax,3),
	RotWallNorm(WallsMax,3),
	Ref1(3), Ref2(3), Ref3(3); \reference frame (unit vectors)
int	Offset(3),		   \offset position to add to vertices (X,Y,Z)
	RotNode(NodesMax,3),	   \rotated vertex coordinates ready to draw
	RotWall(WallsMax,4,3);	   \rotated wall corners, ready to draw

real	FloorVert(FloorsMax,3,3),  \triangular floor section, 3 points, 3D
	FloorNorm(FloorsMax,3),	   \floor section's normal vector (for lighting)
	RotFloorNorm(FloorsMax,3); \rotated normal vector
int	RotFloor(FloorsMax,3,3);   \rotated floor section, ready to draw



proc	Project(P0, P1, P2);	\Project midpoint P2 onto circumscribing sphere
real	P0, P1, P2;		\3D points
real	D2, Mul;
def	Rc = sqrt(sq(1.0) + sq(Phi));	\radius of circumscribed sphere
[D2:= sq(P1(0)-P0(0)) + sq(P1(1)-P0(1)) + sq(P1(2)-P0(2));
Mul:= Rc / sqrt(4.0*sq(Rc) - D2);
P2(0):= (P0(0) + P1(0)) * Mul;
P2(1):= (P0(1) + P1(1)) * Mul;
P2(2):= (P0(2) + P1(2)) * Mul;
];	\Project

\-------------------------------------------------------------------------------

proc	DrawCross(X, Y, C);	\Draw crosshairs
int	X, Y, C;		\coordinates (pixels), color
def	S = 10;
[MovePen(X-S, Y);
DrawLine(X+S, Y, C);
MovePen(X, Y-S);
DrawLine(X, Y+S, C);
];	\DrawCross



proc	DrawMaze;		\Draw the maze (onto Image)
int	F, N, M, W, C, D;
int	Tri(4,3);
int	Quad(5,3);
real	DP;

	proc	DrawWall;	\Draw a wall
	[for C:= 0, 4-1 do
	    for D:= 0, 3-1 do
		Quad(C,D):= RotWall(W,C,D);
	DP:= DotProd(RotWallNorm(W), Light);	\Lambert lighting
	DP:= DP*DP;				\specular highlights
	DP:= DP*DP;
	Quad(4):= $030201 * fix(DP*(85.0-Ambient) + Ambient);
	DrawQuad(Quad);

	Quad(4):= Quad(4) & $FF1F00;		\outline wall with darker color
	MovePen(Quad(0,0), Quad(0,1));
	DrawLine(Quad(1,0), Quad(1,1), Quad(4));
	DrawLine(Quad(2,0), Quad(2,1), Quad(4));
	DrawLine(Quad(3,0), Quad(3,1), Quad(4));
	DrawLine(Quad(0,0), Quad(0,1), Quad(4));
	];

[for W:= 0, Walls-1 do			\for all of the walls...
    if RotWall(W,0,2)>=0 ! RotWall(W,1,2)>=0 ! RotWall(W,2,2)>=0 ! RotWall(W,3,2)>=0
    then DrawWall;			\at least one corner is away from viewer

for F:= 0, Floors-1 do			\for all of the floor sections...
    if RotFloorNorm(F,2) < 0.0 then
	[for C:= 0, 3-1 do		\floor section is facing toward viewer
	    for D:= 0, 3-1 do
		Tri(C,D):= RotFloor(F,C,D);
	DP:= DotProd(RotFloorNorm(F), Light);	\Lambert lighting
	DP:= DP*DP;				\specular highlights
	DP:= DP*DP;
	Tri(3):= $010201 * fix(DP*(95.0-Ambient) + Ambient);
	DrawTriangle(Tri);
	];

for W:= 0, Walls-1 do			\for all of the walls...
    if RotWall(W,0,2)<0 ! RotWall(W,1,2)<0 ! RotWall(W,2,2)<0 ! RotWall(W,3,2)<0
    then DrawWall;			\at least one corner is toward viewer

N:= 0;					\node 0 = entrance
DrawCross(RotNode(N,0), RotNode(N,1), $FFFFFF);
N:= Nodes-1;				\last node = exit
DrawCross(RotNode(N,0), RotNode(N,1), $FF3FFF);
];	\DrawMaze

\-------------------------------------------------------------------------------

proc	Sort(AZ, AW, AN, N);	\Shell sort arrays by decreasing Z coordinates
real	AZ,	\address of array of Z coordinates
	AW,	\address of array of wall coordinates
	AN;	\address of array of surface normals
int	N;	\number of elements in arrays (size)
int	I, J, Gap, JG;
real	T;
[Gap:= N>>1;
while Gap > 0 do
	[for I:= Gap, N-1 do
		[J:= I - Gap;
		loop	[JG:= J + Gap;
			if AZ(J) >= AZ(JG) then quit;
			T:= AZ(J);  AZ(J):= AZ(JG);  AZ(JG):= T;	\swap
			T:= AW(J);  AW(J):= AW(JG);  AW(JG):= T;
			T:= AN(J);  AN(J):= AN(JG);  AN(JG):= T;
			J:= J - Gap;
			if J < 0 then quit;
			];
		];
	Gap:= Gap>>1;
	];
];	\Sort



proc	DepthSort(RWall);	\Sort walls by decreasing depth
real	RWall;
int	W;
real	Depth(WallsMax);
[for W:= 0, Walls-1 do	\Find the Z coordinate (*2) of the center of each wall
	Depth(W):= RWall(W,0,2) + RWall(W,2,2);	\opposite corners
Sort(Depth, RWall, RotWallNorm, Walls);
];	\DepthSort

\-------------------------------------------------------------------------------

proc	RotateLight;	\Rotate light source by moving mouse (with button down)
\Inputs: Light
\Outputs: Light.
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
[CallInt($33, $0B);		\read mouse motion counters
\Mouse X motion rotates object about Y axis
AngY:= Float(-WExtend(CpuReg(2))) / 1500.0;
\Mouse Y motion rotates object about X axis
AngX:= Float( WExtend(CpuReg(3))) / 1500.0;
AngZ:= 0.0;
Rotate(Light, AngX, AngY, AngZ);
];	\RotateLight



proc	RotateMaze;		\Rotate, pan, and zoom maze with the mouse
int	F, N, D, W, C;		\indexes
real	AngX, AngY, AngZ;	\rotation angle about indicated axis (radians)
real	RFloor(FloorsMax,3,3),  \triangular floor section, 3 points, 3D
	RWall(WallsMax,4,3);	\maximum possible walls, 4 corners, 3 dimensions
def	CX=ImW/2, CY=ImH/2;
[AngY:= 0.0;
 AngX:= 0.0;
 AngZ:= 0.0;

CallInt($33, $0003);
case CpuReg(1) & $03 of
  $02:	[	\right button
	CallInt($33, $0B);		\read mouse motion counters
	Offset(0):= Offset(0) + WExtend(CpuReg(2));	\pan motion
	Offset(1):= Offset(1) + WExtend(CpuReg(3));
	];
  $03:	[	\both buttons
	CallInt($33, $0B);		\read mouse motion counters
	Zoom:= Zoom + Float(-WExtend(CpuReg(3)))/10.0;
	]
other	[CallInt($33, $0B);		\read mouse motion counters
	\Mouse X motion rotates object about Y axis
	AngY:= Float(-WExtend(CpuReg(2))) / 500.0;
	\Mouse Y motion rotates object about X axis
	AngX:= Float( WExtend(CpuReg(3))) / 500.0;
	];
\Rotate reference frame according to changes in mouse position.
Rotate(Ref1, AngX, AngY, AngZ);
Rotate(Ref2, AngX, AngY, AngZ);
CrossProd(Ref1, Ref2, Ref3);		\Ref3:= Ref1 x Ref2

\Rotate each vertex in the maze so that it aligns with the reference frame.
\Multiply each point at its initial position (Po) by the rotation matrix to get
\ its rotated position:
\					     				 
\		     	 		     Ref1(0)  Ref1(1)  Ref1(2) 
\	 Px  Py  Pz  =  Pxo	Pyo  Pzo  *  Ref2(0)  Ref2(1)  Ref2(2) 
\		     	 		     Ref3(0)  Ref3(1)  Ref3(2) 
\					     				 
\
for F:= 0, Floors-1 do
  for C:= 0, 3-1 do			\for all the corners...
    for D:= 0, 3-1 do			\for all dimensions (X,Y,Z)...
	RFloor(F,C,D):= FloorVert(F,C,0)*Ref1(D) + FloorVert(F,C,1)*Ref2(D) +
			FloorVert(F,C,2)*Ref3(D);
for F:= 0, Floors-1 do
    for D:= 0, 3-1 do			\for all dimensions (X,Y,Z)...
	RotFloorNorm(F,D):= FloorNorm(F,0)*Ref1(D) + FloorNorm(F,1)*Ref2(D) +
			FloorNorm(F,2)*Ref3(D);
for W:= 0, Walls-1 do
    for D:= 0, 3-1 do			\for all dimensions (X,Y,Z)...
	RotWallNorm(W,D):= WallNorm(W,0)*Ref1(D) + WallNorm(W,1)*Ref2(D) +
			WallNorm(W,2)*Ref3(D);
for N:= 0, Nodes-1 do
    for D:= 0, 3-1 do			\for all dimensions (X,Y,Z)...
	RotVert(N,D):= GeoVert(N,0)*Ref1(D) + GeoVert(N,1)*Ref2(D) +
			   GeoVert(N,2)*Ref3(D);
for W:= 0, Walls-1 do			\for all the walls...
  for C:= 0, 4-1 do			\for all the corners...
    for D:= 0, 3-1 do			\for all dimensions (X,Y,Z)...
	RWall(W,C,D):= WallVert(W,C,0)*Ref1(D) + WallVert(W,C,1)*Ref2(D) +
			   WallVert(W,C,2)*Ref3(D);

DepthSort(RWall);	\sort walls according to depth

\Add pan and zoom
for F:= 0, Floors-1 do
  for C:= 0, 3-1 do			\for all the corners...
       [RotFloor(F,C,0):= fix(RFloor(F,C,0)*Zoom) + Offset(0) + CX;
	RotFloor(F,C,1):= fix(RFloor(F,C,1)*Zoom) + Offset(1) + CY;
	RotFloor(F,C,2):= fix(RFloor(F,C,2));
       ];

for N:= 0, Nodes-1 do
       [RotNode(N,0):= fix(RotVert(N,0)*Zoom) + Offset(0) + CX;
	RotNode(N,1):= fix(RotVert(N,1)*Zoom) + Offset(1) + CY;
	RotNode(N,2):= fix(RotVert(N,2));
       ];

for W:= 0, Walls-1 do
  for C:= 0, 4-1 do			\for all the corners...
       [RotWall(W,C,0):= fix(RWall(W,C,0)*Zoom) + Offset(0) + CX;
	RotWall(W,C,1):= fix(RWall(W,C,1)*Zoom) + Offset(1) + CY;
	RotWall(W,C,2):= fix(RWall(W,C,2));
       ];
];	\RotateMaze

\============================ Make Geodesic Sphere =============================

proc	MakeGeo;	\Make geodesic sphere
int	Node,		\current node number being added to the Geo arrays
	Level;		\current level of recursion



proc	Breakdown(N0, N1, N2);	\Break triangle into four sub-triangles
int	N0, N1, N2;		\node numbers for triangle's vertices
real	P0, P1, P2;		\3D coordinates of vertices of triangle
real	Q0(3), Q1(3), Q2(3);	\mid points of edges of triangle
int	M0, M1, M2;		\node numbers for triangle formed by mid points



proc	InsertNode(N0, N1);
\Insert Node into node list N0, replacing existing entry N1
int	N0, N1;		\original nodes that were connected to each other
int	I;
[for I:= 0, GeoNodeInx(N0)-1 do
    if GeoNode(N0,I) = N1 then		\search N0's list for N1
	[GeoNode(N0,I):= Node; return];	\update list with new Node
Fatal("Node missing");
];	\InsertNode



func	AddNode(N0, N1, Q);	\Add point Q to Geo arrays, if not already there
int	N0, N1;			\Returns number of new node or existing node
real	Q;
int	N, D;
[for N:= 0, Node-1 do
	if sq(GeoVert(N,0)-Q(0)) + sq(GeoVert(N,1)-Q(1)) + sq(GeoVert(N,2)-Q(2))
		< 0.001 then return N;	\return if point is already in Geo array

if Node < Nodes then
	[for D:= 0, 3-1 do GeoVert(Node,D):= Q(D);	\add point to Geo arrays

	GeoNode(Node,0):= N0;	\insert new node between existing nodes N0 & N1
	GeoNode(Node,1):= N1;
	GeoNodeInx(Node):= 2;
	InsertNode(N0, N1);
	InsertNode(N1, N0);
	Node:= Node+1;
	]
else	Fatal("Overflow");
return Node-1;
];	\AddNode



int	I, D;
[				\Breakdown
\           P0
\           /\
\       Q0 /__\ \Q2
\         /`  /\
\        /__`/__\   
\     P1    Q1    P2

P0:= GeoVert(N0);		\get points corresponding to node #'s
P1:= GeoVert(N1);
P2:= GeoVert(N2);

if Level = Levels then
	[for D:= 0, 3-1 do
		[FloorVert(Floors,0,D):= P0(D);
		 FloorVert(Floors,1,D):= P1(D);
		 FloorVert(Floors,2,D):= P2(D);
		];
	VectSub(P1, P0, Q1);	\norm = p2-p0 x p1-p0
	VectSub(P2, P0, Q2);
	CrossProd(Q2, Q1, FloorNorm(Floors));
	Unitize(FloorNorm(Floors));

	Floors:= Floors+1;
	];

Level:= Level+1;
if Level <= Levels then
	[\given 3 points P, generate 3 more points Q
	Project(P0, P1, Q0);		\generate new point Q0
	M0:= AddNode(N0, N1, Q0);	\add Q0 to Geo arrays and get its node #
	Project(P1, P2, Q1);
	M1:= AddNode(N1, N2, Q1);
	Project(P2, P0, Q2);
	M2:= AddNode(N2, N0, Q2);

	I:= GeoNodeInx(M0);		\make node connections
	GeoNode(M0,I):= M1;		\(gets executed twice for each point
	GeoNode(M0,I+1):= M2;		\ because triangles share common edges)
	GeoNodeInx(M0):= I+2;

	I:= GeoNodeInx(M1);
	GeoNode(M1,I):= M0;
	GeoNode(M1,I+1):= M2;
	GeoNodeInx(M1):= I+2;

	I:= GeoNodeInx(M2);
	GeoNode(M2,I):= M0;
	GeoNode(M2,I+1):= M1;
	GeoNodeInx(M2):= I+2;

	Breakdown(M0, M1, M2);		\similiarly breakdown each sub triangle
	Breakdown(N0, M0, M2);		\ using recursion
	Breakdown(M0, N1, M1);
	Breakdown(M2, M1, N2);
	];
Level:= Level-1;
];	\Breakdown



int	IcosFace, IcosNode, I, D, N, F;
real	IcosVert;	\coordinates of vertexes

[	\MakeGeo
\coordinates of the vertices of an icosahedron
\	   X    Y    Z
IcosVert:=			\Vertex
      [ [ 0.0, Phi, 1.0],	\0
	[ 0.0, Phi,-1.0],	\1
	[ 0.0,-Phi, 1.0],	\2
	[ 0.0,-Phi,-1.0],	\3
	[ Phi, 1.0, 0.0],	\4
	[ Phi,-1.0, 0.0],	\5
	[-Phi, 1.0, 0.0],	\6
	[-Phi,-1.0, 0.0],	\7
	[ 1.0, 0.0, Phi],	\8
	[ 1.0, 0.0,-Phi],	\9
	[-1.0, 0.0, Phi],	\10
	[-1.0, 0.0,-Phi]];	\11

IcosNode:=			\node connections
      [ [1, 4, 8,10, 6],	\0
	[0, 6,11, 9, 4],	\1
	[3, 7,10, 8, 5],	\2
	[2, 5, 9,11, 7],	\3
	[0, 1, 9, 5, 8],	\4
	[2, 8, 9, 4, 3],	\5
	[0,10, 7,11, 1],	\6
	[2, 3,11, 6,10],	\7
	[0, 4, 5, 2,10],	\8
	[1,11, 3, 5, 4],	\9
	[0, 8, 2, 7, 6],	\10
	[1, 6, 7, 3, 9] ];	\11

\Vertices for each face of icosahedron
IcosFace:=
      [ [0,1,4],	[0,4,8],	[0,8,10],	[0,10,6],	[0,6,1],
	[1,9,4],	[4,5,8],	[8,2,10],	[10,7,6],	[6,11,1],	
	[4,9,5],	[8,5,2],	[10,2,7],	[6,7,11],	[1,11,9],
	[9,3,5],	[5,3,2],	[2,3,7],	[7,3,11],	[11,3,9]];

for N:= 0, Nodes-1 do GeoNodeInx(N):= 0;	\initialize indexes (for safety)

for D:= 0, 3-1 do
    for N:= 0, 12-1 do
	GeoVert(N,D):= IcosVert(N,D);

for N:= 0, 12-1 do
	[for I:= 0, 5-1 do
		GeoNode(N,I):= IcosNode(N,I);
	GeoNodeInx(N):= 5;
	];
Node:= 12;
Floors:= 0;
Level:= 0;
for F:= 0, 20-1 do
    Breakdown(IcosFace(F,0), IcosFace(F,1), IcosFace(F,2));
];	\MakeGeo

\===============================================================================

proc	Center(V0, V1, V2, C);	\Return the center (C) of three 3D vectors
real	V0, V1, V2, C;		\C:= (V0+V1+V2)/3
[C(0):= (V0(0) + V1(0) + V2(0)) / 3.0;
 C(1):= (V0(1) + V1(1) + V2(1)) / 3.0;
 C(2):= (V0(2) + V1(2) + V2(2)) / 3.0;
];	\Center



proc	MakeWalls;		\Erect walls between unconneced nodes
int	W, I, J, Cor, N0, N1, N2; \Outputs WallVert array and Walls
real	P0, P1, P2,
	Q1(3), Q2(3),
	Mag;
[Walls:= 0;
for N0:= 0, Nodes-1 do				\for all the nodes...
    for I:= 0, GeoNodeInx(N0)-1 do		\for its all connecting nodes...
	[N1:= GeoNode(N0,I);
	if not GeoPath(N0,I) & N0<N1 then	\if no path then put up wall
	    [Cor:= 0;				\start with a lower corner
	    for J:= 0, GeoNodeInx(N0)-1 do	\search for adjacent nodes
	      if J # I then
		[N2:= GeoNode(N0,J);
		P0:= GeoVert(N0);
		P1:= GeoVert(N1);
		P2:= GeoVert(N2);
		VectSub(P1, P0, Q1);
		VectSub(P2, P0, Q2);
		if DotProd(Q1, Q2) > 0.0 then	\node N2 is adjacent to N1
			[Center(P0, P1, P2, WallVert(Walls,Cor)); \triangle
			Project(P0, P1, Q1);	\elevate wall above ridge
			Mag:= VectMag(Q1) / VectMag(WallVert(Walls,Cor));
			VectMul(WallVert(Walls,Cor), Mag, WallVert(Walls,Cor));
			VectMul(WallVert(Walls,Cor), 1.15,	 \upper corners
				WallVert(Walls, if Cor=0 then 3 else 2));
			if Cor=0 then Cor:= 1;	\repeat for other lower corner
			];
		];
	    Walls:= Walls+1;			\count walls actually erected
	    ];
	];

for W:= 0, Walls-1 do				\gen normal vectors for lighting
	[CrossProd(WallVert(W,0), WallVert(W,1), WallNorm(W));
	Unitize(WallNorm(W));			\wall can face either direction
	];
];	\MakeWalls

\-------------------------------------------------------------------------------

proc	SolveMaze;		\Use dead-end-filling to show path thru maze
int	N, M, I, J, C;
[Solve:= false;			\assume it's done solving
for N:= 1, Nodes-2 do		\scan through entire maze except entrance & exit
	[C:= 0;			\count number of paths leading to a node
	for I:= 0, GeoNodeInx(N)-1 do
		if GeoPath(N,I) then [C:= C+1;  J:= I];
	if C = 1 then		\if only one path then remove it
		[GeoPath(N,J):= false;
		M:= GeoNode(N,J);
		\remove corresponding path in the node that's connected to
		for I:= 0, GeoNodeInx(M)-1 do
			if GeoNode(M,I) = N then GeoPath(M,I):= false;
		Solve:= true;	\nope, still working on solution
		];
	];
if Solve then MakeWalls;	\handle the newly erected walls
];	\SolveMaze

\-------------------------------------------------------------------------------

proc	MakeMaze;	\Randomly generate maze and set up array of coordinates
int	Connected(NodesMax);



proc	ConnectFrom(N);		\Connect nodes starting from node N
int	N;
int	Dir, Dir0, M, I;
[Connected(N):= true;		\mark current node as connected
Dir:= Ran(GeoNodeInx(N));	\randomly choose a direction (0 to 5, or 0 to 6)
Dir0:= Dir;			\save this initial direction
repeat	M:= GeoNode(N,Dir);	\node M is next to node N
	if not Connected(M) then		\try to connect to node at Dir
		[GeoPath(N,Dir):= true;		\form path between nodes N & M
		for I:= 0, GeoNodeInx(M)-1 do
			if GeoNode(M,I) = N then GeoPath(M,I):= true;
		ConnectFrom(M);			\recurse
		];
	Dir:= rem((Dir+1) / GeoNodeInx(N));	\next direction for current node
until	Dir = Dir0;				\back where it started?
];	\ConnectFrom



proc	GenMaze;		\Generate random maze
int	N, D;
[for N:= 0, Nodes-1 do
	[for D:= 0, 6-1 do GeoPath(N,D):= false; \start with no pathways
	Connected(N):= false;			 \and all nodes disconnected
	];
ConnectFrom(Ran(Nodes));			\randomly pick a starting node
];	\GenMaze



int	Edges, L;

[	\MakeMaze
Nodes:= 12;
Edges:= 30;
for L:= 0, Levels-1 do
	[Nodes:= Nodes + Edges;
	Edges:= Edges * 4;
	];
MakeGeo;
GenMaze;
MakeWalls;
];	\MakeMaze

\===============================================================================

proc	Reset;			\Initialize variables to starting configuration
[Levels:= 2;
Zoom:= 170.0;			\magnification factor
Offset(0):= 0;   Offset(1):= 0;   Offset(2):= 0;

Ref1(0):= 1.0;   Ref1(1):= 0.0;   Ref1(2):= 0.0;	\tilt reference frame 30
Ref2(0):= 0.0;   Ref2(1):= sqrt(3.0)/2.0;   Ref2(2):= -0.5;

Ambient:= 45.0;
Light(0):= 0.0; Light(1):= 0.0; Light(2):= -1.0; \behind and in line with viewer

MakeMaze;

Solve:= false;
];	\Reset



proc	NumOut(Line, Num);	\Display 3-digit right-justified positive number
int	Line, Num;		\ at the beginning of the specified Line
[Cursor(0, Line);
if Num < 100 then ChOut(6, ^ );
if Num < 10  then ChOut(6, ^ );
IntOut(6, Num);
];	\NumOut

\-------------------------------------------------------------------------------

[	\MoveMaze
FPS:= 0;
Reset;

Cursor(0, 0);			\(device 6 always works for VESA graphic modes)
Text(6, "Mouse rotates maze
Right button pans
Left button moves light
Both buttons zoom

    Levels   [^Y] [^X]



    Frames/second

    Ambient  [-] [+]

    Solve    [S]

    Reset    [Enter]

    Exit     [Esc]");

loop	[NumOut(5, Levels);

	CallInt($1A, $0200);	\get current time in seconds
	Sec:= CpuReg(3)>>8 & $FF;
	if Sec = SecOld then	\if same as before then
		FPS:= FPS+1	\ bump FPS counter
	else	[SecOld:= Sec;	\save old seconds
		NumOut(9, FPS);	\display FPS counter (once per second)
		FPS:= 0;	\reset FPS counter
		];

	NumOut(11, fix(Ambient));

	if Solve then SolveMaze;

	CallInt($33, $0003);
	if (CpuReg(1)&$03) = $01 \left\ then RotateLight;

	RotateMaze;

	EraseImage;
	DrawMaze;
	Paint(ScrW-ImW, 0, ImW, ImH, Image, ImW);     \copy Image onto screen

	if KeyHit then
		case GetKey of
		  ^-:	[Ambient:= Ambient - 3.0;
			if Ambient < 0.0 then Ambient:= 0.0];
		  ^+:	[Ambient:= Ambient + 3.0;
			if Ambient > 255.0 then Ambient:= 255.0];
		  -UpArrow:
			if Levels < LevelsMax-1 then
				[Levels:= Levels+1;  MakeMaze];
		  -DnArrow:
			if Levels > 0 then
				[Levels:= Levels-1;  MakeMaze];
		  ^S, ^s:	Solve:= true;
		  CR:	Reset;
		  Esc:	quit
		other	ChOut(0, Bel);
	];
];	\MoveMaze

\###############################################################################

[	\Main
CpuReg:= GetReg;		\get address of array to access CPU registers
SetVid(VidMode);		\(in WinXP this affects mouse motion counters)
if not OpenMouse then Fatal("A mouse is required");
MoveMaze;
SetVid(3);			\restore normal text mode (for DOS)
];	\Main
