\RatsMaze	27-Aug-2009	Loren Blaney	loren.blaney@idcomm.com
\Use the mouse to find your way out of a maze
\Compile with 16-bit XPL (xx.bat)

\Each cell in a randomly generated maze is converted into 4 cell blocks for the
\ ray caster. Each of these cell blocks is divided into a 64x64 sub-cell
\ footprint.

inc	c:\cxpl\codesi;		\intrinsic code declarations

def	Cols=12, Rows=8;	\dimensions of maze (cells)
def	Barrier= 20;		\how near the viewer can get to a wall, subcells

int	CpuReg,			\address of CPU register array (from GetReg)
	MouseDx, MouseDy,	\amount mouse moved since last GetMouseMove call
	Vx, Vy;			\viewer's coordinates (subcells)
real	Vd;			\viewer's direction (radians: 0=east pi/2=south)
int	CellBlk(2*Cols+2, 2*Rows+2); \contains 3D blocks that form the maze

char	SpWall(14, 2+64*64);	\wall panel images (sprites)
seg char Image(1);		\video screen buffer

def	Pi = 3.141592654,
	PiOver2 = Pi/2.0;	\90 degrees



proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
begin
SetVid(3);			\restore normal text mode
Text(0, Msg);
CrLf(0);
OpenI(0);			\discard any keystrokes
if GetErr then [];		\discard any pending runtime errors
exit;
end;	\Exit



func	Trunc(X);		\Convert real to integer by truncation
real	X;
return Fix(X-0.5);

\=============================== MOUSE ROUTINES ================================

func	OpenMouse;		\Initialize mouse; return 'false' if it fails
begin				\pointer is set to center of screen but is hidden
CpuReg(0):= $3533;		\make sure mouse vector ($33) points to something
SoftInt($21);
if CpuReg(1)=0 & CpuReg(11)=0 then return false;

CpuReg(0):= $0000;		\reset mouse
SoftInt($33);
return CpuReg(0)		\return 'false' if failure
end;	\OpenMouse



proc	GetMouseMove;		\Get amount mouse moved since last call
begin
CpuReg(0):= $000B;
SoftInt($33);
MouseDx:= CpuReg(2);
MouseDy:= CpuReg(3);
end;	\GetMouseMove

\================================ IMAGE LOADING ================================

proc	LoadWalls;	\Load sprite images from .BMP file and set up color regs
int	Hand,		\file handle
	W, H,		\dimensions of entire image in .BMP file (pixels)
	R, G, B,	\red, green, blue
	X, Y,		\screen coordinates (pixels)
	Y320,		\Y*320 (for efficiency)
	I, T;		\index and temporary scratch
seg char Screen(1);	\temporary memory area for loading images



proc	LoadWall(X0, Y0, W, H, Sp);	\Load a sprite data array from Screen
int	X0, Y0,		\coordinates of sprite in Screen (pixels)
	W, H;		\width and height of sprite (pixels)
char	Sp;		\sprite to load
int	I, X, Y;
begin
Sp(0):= W;
Sp(1):= H;
I:= 2;
for Y:= Y0, Y0+H-1 do
    for X:= X0, X0+W-1 do
	begin
	Sp(I):= Screen(0, X+Y*320);
	I:= I + 1;
	end;
end;	\LoadWall



begin	\LoadWalls
Screen(0):= Malloc(320/16*200);	\temporary 64000-byte hunk of memory

\Read in a 256-color .BMP file
Trap(false);			\don't trap "file not found" error; we'll do it
Hand:= FOpen("WALLS.BMP", 0);	\open file for input
if GetErr \#0\ then Exit("WALLS.BMP not found");
Trap(true);
FSet(Hand, ^I);			\set device 3 to file handle
OpenI(3);

for I:= 0, 17 do T:= Chin(3);	\skip unused header info
W:= Chin(3) + Chin(3)<<8;	\0..32767 (is ample range)
for I:= 0, 2-1 do T:= Chin(3);	\skip
H:= Chin(3) + Chin(3)<<8;	\0..32767 (is ample range)
for I:= 24, 53 do T:= Chin(3);	\skip

port($3C8):= 0;			\set color registers
for I:= 0, 255 do
	begin
	B:= Chin(3)>>2;
	G:= Chin(3)>>2;
	R:= Chin(3)>>2;
	T:= Chin(3);
	port($3C9):= R;
	port($3C9):= G;
	port($3C9):= B;
	end;

\Load .BMP image into Screen array
for Y:= -(H-1), 0 do		\.BMP files are upside down (!)
	begin
	Y320:= 320 * -Y;
	for X:= 0, W-1 do
		Screen(0, X+Y320):= Chin(3);
	end;
FClose(Hand);			\close handle (so it could be used again)

for I:= 0, 4 do
	begin
	LoadWall(I*64,  0, 64, 64, addr SpWall(I,0));
	LoadWall(I*64, 64, 64, 64, addr SpWall(I+5,0));
	end;
for I:= 0, 3 do
	LoadWall(I*64,128, 64, 64, addr SpWall(I+10,0));

Release(Screen(0));		\(for neatness, so memory can be reused)
end;	\LoadWalls

\================================= RAY CASTER ==================================

proc	DrawWalls;	\Draw a series of wall panels forming the rooms
int	X, Y,		\coordinates in Image array (pixels)
	Y320,		\Y*320 (for efficiency)
	CBx, CBy,	\cell boundary coordinates
	Wy,		\coordinates in wall (Y direction is vertical)
	Wx1,		\X offset into a wall panel
	WP,		\wall panel code (0..$F)
	Dist1,		\modified distance, used to eliminate fish-eye effect
	Color;		\color of pixel that gets displayed
real	Wx,		\coordinates in wall (Y direction is vertical)
	Vx1, Vy1,	\coordinates of start of ray (hops from cell to cell)
	VA,		\view angle of ray (in radians) relative to viewer (Vd)
	Dv, Dh,		\distances to vertical & horizontal cell boundaries
	Dist,		\distance from viewer to point where ray hits wall
	SinA, CosA;	\sine and cosine of view angle (for efficiency)
int	Face, FaceV, FaceH;
def	N=12, W=8, S=4, E=0;

begin
for X:= 0, 320-1 do	\raster fill Image (up & down instead of left & right)
	begin
	\Sweep projected ray from -30 to +30 degrees (sweep from left to right)
	VA:= Atan2(Float(X-160), 277.2);

	\Given viewer's position and ray's direction find Wx and Dist:
	CosA:= Cos(VA+Vd);
	SinA:= Sin(VA+Vd);
	Vx1:= Float(Vx);		\working copy of ray's starting position
	Vy1:= Float(Vy);
	repeat	begin			\loop thru cells until ray hits a wall
		if CosA >= 0.0 then			\ray heading toward East
			begin
			CBx:= (Trunc(Vx1)+64) & ~$3F;	\cell's eastern boundary
			FaceV:= W;
			end
		else	begin
			CBx:= (Trunc(Vx1) & ~$3F) - 1;	\cell's western boundary
			FaceV:= E;
			end;
		Dv:= (Float(CBx)-Vx1)/CosA;		\distance to boundary

		if SinA >= 0.0 then			\ray heading Southward
			begin
			CBy:= (Trunc(Vy1)+64) & ~$3F;	\cell's south boundary
			FaceH:= N;
			end
		else	begin
			CBy:= (Trunc(Vy1) & ~$3F) - 1;	\cell's north boundary
			FaceH:= S;
			end;
		Dh:= (Float(CBy)-Vy1)/SinA;		\distance to boundary

		if Dv <= Dh then		\use the shorter distance
			begin			\toward vertical cell boundary
			Wx:= Vy1 + Dv*SinA;	\intercept on vertical boundary
			Vx1:= Float(CBx);	\continue ray thru next cell
			Vy1:= Wx;
			Face:= FaceV;
			end
		else	begin			\toward horizontal cell boundary
			Wx:= Vx1 + Dh*CosA;	\intercept on horz. boundary
			Vx1:= Wx;		\continue ray thru next cell
			Vy1:= Float(CBy);
			Face:= FaceH;
			end;
		if Vx1<0.0 ! Vy1<0.0 then WP:= $5555	\wall panel of next cell
		else WP:= CellBlk(Trunc(Vx1)>>6, Trunc(Vy1)>>6);
		end;
	until	WP \#0\;			\ray struck a wall

	Dv:= Vx1 - Float(Vx);			\distance from viewer to wall
	Dh:= Vy1 - Float(Vy);
	Dist:= Sqrt(Dv*Dv + Dh*Dh);
	Dist1:= Fix(Dist * Cos(VA)) /2;		\eliminate fish-eye effect
						\ (use parallel projection)
	Wx1:= Fix(Wx) & $3F;			\horz. position in wall panel
	WP:= WP>>Face & $000F;

	Y320:= 0;			\set up column of pixels to be displayed
	for Y:= 0, 200-1 do
		begin			\speed-critical code: 64000 per frame
		Wy:= ((Y-100)*Dist1)/128 + 32;	\256 = height scale factor
		case of				\100=200/2; 32=64/2; (centering)
		  Wy<0:  Color:= 7;				\ceiling
		  Wy>63: Color:= $A4				\floor
		other	 Color:= SpWall(WP-1, 2+Wx1+Wy<<6);	\wall panel
		Image(0, X+Y320):= Color;
		Y320:= Y320 + 320;
		end;
	end;
end;	\DrawWalls

\-------------------------------------------------------------------------------

proc	MoveViewer;
\Move viewer's position (Vx,Vy) and direction (Vd) using the mouse
\ -Y axis moves forward and back
\  X axis moves clockwise and counter-clockwise
real	A, D, DMax;
int	Vx6, Vy6,
	CBx, CBy;	\cell boundaries
begin
GetMouseMove;				\get change in mouse position
A:= Float(MouseDx) / 1000.0;		\X-axis motion is change in direction
Vd:= Vd + A;
D:= -Float(MouseDy/3);			\Y-axis motion moves along curent dir
DMax:= Float(Barrier-1);		\speed limit prevents moving thru walls
if D > DMax then D:= DMax;
if D < -DMax then D:= -DMax;

\Move viewer and limit how close he can get to a cell containing a block
Vx:= Vx + Fix(D*Cos(Vd));		\make tenative East-West move
Vx6:= Vx>>6;
Vy6:= Vy>>6;

CBx:= (Vx+64) & ~$3F;
if CellBlk(Vx6+1, Vy6) \#0\ then	\cell to the East contains a block
	if Vx > CBx-Barrier then Vx:= CBx-Barrier;
CBy:= (Vy+64) & ~$3F;
if CellBlk(Vx6+1, Vy6+1) \#0\ then	\cell to the SE contains a block
	if Vx>CBx-Barrier & Vy>CBy-Barrier then Vx:= CBx-Barrier;
CBy:= Vy & ~$3F;
if CellBlk(Vx6+1, Vy6-1) \#0\ then	\cell to the NE contains a block
	if Vx>CBx-Barrier & Vy<CBy+Barrier then Vx:= CBx-Barrier;

CBx:= Vx & ~$3F;
if CellBlk(Vx6-1, Vy6) \#0\ then	\cell to the West contains a block
	if Vx < CBx+Barrier then Vx:= CBx+Barrier;
CBy:= (Vy+64) & ~$3F;
if CellBlk(Vx6-1, Vy6+1) \#0\ then	\cell to the SW contains a block
	if Vx<CBx+Barrier & Vy>CBy-Barrier then Vx:= CBx+Barrier;
CBy:= Vy & ~$3F;
if CellBlk(Vx6-1, Vy6-1) \#0\ then	\cell to the NW contains a block
	if Vx<CBx+Barrier & Vy<CBy+Barrier then Vx:= CBx+Barrier;

Vy:= Vy + Fix(D*Sin(Vd));
Vx6:= Vx>>6;
Vy6:= Vy>>6;

CBy:= (Vy+64) & ~$3F;
if CellBlk(Vx6, Vy6+1) \#0\ then	\cell to the South contains a block
	if Vy > CBy-Barrier then Vy:= CBy-Barrier;
CBx:= (Vx+64) & ~$3F;
if CellBlk(Vx6+1, Vy6+1) \#0\ then	\cell to the SE contains a block
	if Vy>CBy-Barrier & Vx>CBx-Barrier then Vy:= CBy-Barrier;
CBx:= Vx & ~$3F;
if CellBlk(Vx6-1, Vy6+1) \#0\ then	\cell to the SW contains a block
	if Vy>CBy-Barrier & Vx<CBx+Barrier then Vy:= CBy-Barrier;

CBy:= Vy & ~$3F;
if CellBlk(Vx6, Vy6-1) \#0\ then	\cell to the North contains a block
	if Vy < CBy+Barrier then Vy:= CBy+Barrier;
CBx:= (Vx+64) & ~$3F;
if CellBlk(Vx6+1, Vy6-1) \#0\ then	\cell to the NE contains a block
	if Vy<CBy+Barrier & Vx>CBx-Barrier then Vy:= CBy+Barrier;
CBx:= Vx & ~$3F;
if CellBlk(Vx6-1, Vy6-1) \#0\ then	\cell to the NW contains a block
	if Vy<CBy+Barrier & Vx<CBx+Barrier then Vy:= CBy+Barrier;
end;	\MoveViewer



proc	DrawBlock(X, Y, C);
int	X, Y, C;	\coordinates * Size, color
int	I, J;
def	Size = 3;
def	X0 = 320 - Cols*2*Size - 8,	\position on screen (pixels)
	Y0 = 5;
begin
for J:= Y0+Y*Size, Y0+Y*Size+Size-1 do
    for I:= X0+X*Size, X0+X*Size+Size-1 do
	Image(0, I+J*320):= C;
end;	\DrawBlock



proc	DrawMap;
int	X, Y;
begin
for Y:= 0, Rows*2 do
    for X:= 0, Cols*2 do
	if CellBlk(X, Y) then DrawBlock(X, Y, 8 \gray\);
DrawBlock(Vx>>6, Vy>>6, $F);
end;	\DrawMap



proc	CastRays;
begin
\CellBlk is an array of cells describing the floor plan.
\ Each cell is either open (=0) or occupied by a block.
\ Each block has 4 walls.
\ Each wall is specified by a hex digit 1..F.
\ The direction a wall faces is specified by the order of the digits: N,W,S,E.
\ For example, the block $1234 has wall image number 2 facing West.

Vx:= 64*2*(Cols-1) + 64 + Barrier;	\start viewer in south-east corner
Vy:= 64*2*(Rows-1) + 64 + Barrier;
Vd:= -PiOver2/2.0*3.0;			\face northwest

GetMouseMove;				\initialize mouse's starting position
repeat	MoveViewer;
	DrawWalls;			\draw into Image array
	DrawMap;
	repeat until port($3DA) & $08;	\wait for vertical blank
	\Copy Image to screen (takes about 5ms--outruns video beam--no glitches)
	Blit(Image(0), 0, $A000, 0, 64*1000);
until ChkKey;
Exit(" ");				\terminate program when a key is hit
end;	\CastRays

\================================ GENERATE MAZE ================================

proc	GenMaze;			\Random Maze Generator
int	Cell(Cols+1, Rows+1, 3);	\cells (plus right and bottom borders)
def	LeftWall, Ceiling, Connected;	\attributes of each cell (= 0, 1 and 2)



proc	ConnectFrom(X, Y);		\Connect cells starting from cell X,Y
int	X, Y;
int	Dir, Dir0;
begin
Cell(X, Y, Connected):= true;		\mark current cell as connected
Dir:= Ran(4);				\randomly choose a direction
Dir0:= Dir;				\save this initial direction
repeat	case Dir of			\try to connect to cell at Dir
	  0: if X+1<Cols & not Cell(X+1, Y, Connected) then	\go right
		[Cell(X+1, Y, LeftWall):= false; ConnectFrom(X+1, Y)];
	  1: if Y+1<Rows & not Cell(X, Y+1, Connected) then	\go down
		[Cell(X, Y+1, Ceiling):= false;  ConnectFrom(X, Y+1)];
	  2: if X-1>=0 & not Cell(X-1, Y, Connected) then	\go left
		[Cell(X, Y, LeftWall):= false;   ConnectFrom(X-1, Y)];
	  3: if Y-1>=0 & not Cell(X, Y-1, Connected) then	\go up
		[Cell(X, Y, Ceiling):= false;    ConnectFrom(X, Y-1)]
	other	[];			\(never occurs)
	Dir:= Dir+1 & $03;		\next direction
until	Dir = Dir0;
end;	\ConnectFrom



int	X, Y;
def	Blk = $7666;
begin	\GenMaze
for Y:= 0, Rows do
    for X:= 0, Cols do
	begin
	Cell(X, Y, LeftWall):= true;	\start with all walls and
	Cell(X, Y, Ceiling):= true;	\ ceilings in place
	Cell(X, Y, Connected):= false;	\ and all cells disconnected
	end;
Cell(0, 0, LeftWall):= false;		\make doorway

ConnectFrom(Ran(Cols), Ran(Rows));	\randomly pick a starting cell

for Y:= 0, Rows do			\display the maze
	begin
	CrLf(0);
	for X:= 0, Cols do
		begin
		CellBlk(2*X, 2*Y):= Blk;
		CellBlk(2*X+1, 2*Y):= if Cell(X, Y, Ceiling) then Blk else 0;
		CellBlk(2*X, 2*Y+1):= if Cell(X, Y, LeftWall) then Blk else 0;
		CellBlk(2*X+1, 2*Y+1):= 0;
		end;
	end;
end;	\GenMaze

\===============================================================================

begin	\Main
Image(0):= Malloc(320/16*200);		\64000 bytes

CpuReg:= GetReg;
if not OpenMouse then Exit("A mouse is required");

SetVid($13);				\320x200 with 256 colors
LoadWalls;
GenMaze;
CastRays;
end;	\Main
